@@ -5,6 +5,7 @@
  * Copyright (c) 2004-2005 Holger Ohmacht
  * Copyright (c) 2005      Anton Altaparmakov
  * Copyright (c) 2007      Yura Pakhuchiy
+ * Copyright (c) 2013      Jean-Pierre Andre
  *
  * This utility will recover deleted files from an NTFS volume.
  *
@@ -68,12 +69,18 @@
 #ifdef HAVE_UTIME_H
 #include <utime.h>
 #endif
+#ifdef HAVE_REGEX_H
 #include <regex.h>
+#endif
 
 #if !defined(REG_NOERROR) || (REG_NOERROR != 0)
 #define REG_NOERROR 0
 #endif
 
+#ifndef REG_NOMATCH
+#define REG_NOMATCH 1
+#endif
+
 #include "ntfsundelete.h"
 #include "bootsect.h"
 #include "mft.h"
@@ -86,6 +93,14 @@
 #include "ntfstime.h"
 /* #include "version.h" */
 #include "logging.h"
+#include "misc.h"
+
+#ifdef HAVE_WINDOWS_H
+/*
+ *		Replacements for functions which do not exist on Windows
+ */
+#define ftruncate(fd, size) ntfs_win32_ftruncate(fd, size)
+#endif
 
 static const char *EXEC_NAME = "ntfsundelete";
 static const char *MFTFILE   = "mft";
@@ -105,6 +120,157 @@ static short	avoid_duplicate_printing;	/* Flag  No duplicate printing of file in
 static range	*ranges;			/* Array containing all Inode-Ranges for undelete */
 static long	nr_entries;			/* Number of range entries */
 
+#ifdef HAVE_WINDOWS_H
+/*
+ *		Replacement for strftime() on Windows
+ *
+ *	strftime() on Windows uses format codes different from those
+ *	defined in C99 sect. 7.23.3.5
+ *	Use snprintf() instead.
+ */
+static int win32_strftime(char *buffer, int size, const char *format,
+					const struct tm *ptm)
+{
+	int ret;
+
+	if (!strcmp(format, "%F %R"))
+		ret = snprintf(buffer, size, "%4d-%02d-%02d %02d:%02d",
+			ptm->tm_year + 1900, ptm->tm_mon + 1, ptm->tm_mday,
+			ptm->tm_hour, ptm->tm_min);
+	else
+		ret = snprintf(buffer, size, "%4d-%02d-%02d",
+			ptm->tm_year + 1900, ptm->tm_mon + 1, ptm->tm_mday);
+	return (ret);
+}
+#define strftime(buf, sz, fmt, ptm) win32_strftime(buf, sz, fmt, ptm)
+#endif
+
+#ifndef HAVE_REGEX_H
+
+/*
+ *		Pattern matching routing for systems with no regex.
+ */
+
+typedef struct REGEX {
+	ntfschar *upcase;
+	u32 upcase_len;
+	int flags;
+	int pattern_len;
+	ntfschar pattern[1];
+} *regex_t;
+
+enum { REG_NOSUB = 1, REG_ICASE = 2 };
+
+static BOOL patmatch(regex_t *re, const ntfschar *f, int flen,
+			const ntfschar *p, int plen, BOOL dot)
+{
+	regex_t pre;
+	BOOL ok;
+	BOOL anyextens;
+	int i;
+	unsigned int c;
+
+	pre = *re;
+	if (pre->flags & REG_ICASE) {
+		while ((flen > 0) && (plen > 0)
+		    && ((*f == *p)
+			|| (*p == const_cpu_to_le16('?'))
+			|| ((c = le16_to_cpu(*f)) < pre->upcase_len
+				? pre->upcase[c] : c) == *p)) {
+			flen--;
+			if (*f++ == const_cpu_to_le16('.'))
+				dot = TRUE;
+			plen--;
+			p++;
+		}
+	} else {
+		while ((flen > 0) && (plen > 0)
+		    && ((*f == *p) || (*p == const_cpu_to_le16('?')))) {
+			flen--;
+			if (*f++ == const_cpu_to_le16('.'))
+				dot = TRUE;
+			plen--;
+			p++;
+		}
+	}
+	if ((flen <= 0) && (plen <= 0))
+		ok = TRUE;
+	else {
+		ok = FALSE;
+		plen--;
+		if (*p++ == const_cpu_to_le16('*')) {
+			/* special case "*.*" requires the end or a dot */
+			anyextens = FALSE;
+			if ((plen == 2)
+			    && (p[0] == const_cpu_to_le16('.'))
+			    && (p[1] == const_cpu_to_le16('*'))
+			    && !dot) {
+				for (i=0; (i<flen) && !anyextens; i++)
+					if (f[i] == const_cpu_to_le16('.'))
+						anyextens = TRUE;
+			}
+			if (!plen || anyextens)
+				ok = TRUE;
+			else
+				while ((flen > 0) && !ok)
+					if (patmatch(re,f,flen,p,plen,dot))
+						ok = TRUE;
+					else {
+						flen--;
+						f++;
+					}
+		}
+	}
+	return (ok);
+}
+
+static int regcomp(regex_t *re, const char *pattern, int flags)
+{
+	regex_t pre;
+	ntfschar *rp;
+	ntfschar *p;
+	unsigned int c;
+	int lth;
+	int i;
+
+	pre = (regex_t)malloc(sizeof(struct REGEX)
+			+ strlen(pattern)*sizeof(ntfschar));
+	*re = pre;
+	if (pre) {
+		pre->flags = flags;
+		pre->upcase_len = 0;
+		rp = pre->pattern;
+		lth = ntfs_mbstoucs(pattern, &rp);
+		pre->pattern_len = lth;
+		p = pre->pattern;
+		if (flags & REG_ICASE) {
+			for (i=0; i<lth; i++) {
+				c = le16_to_cpu(*p);
+				if (c < pre->upcase_len)
+					*p = pre->upcase[c];
+				p++;
+			}
+		}
+	}
+	return (*re && (lth > 0) ? 0 : -1);
+}
+
+static int regexec(regex_t *re, const ntfschar *uname, int len,
+		char *q __attribute__((unused)), int r __attribute__((unused)))
+{
+	BOOL m;
+
+	m = patmatch(re, uname, len, (*re)->pattern, (*re)->pattern_len, FALSE);
+	return (m ? REG_NOERROR : REG_NOMATCH);
+}
+
+static void regfree(regex_t *re)
+{
+	free(*re);
+}
+
+#endif
+
 /**
  * parse_inode_arg - parses the inode expression
  *
@@ -115,7 +281,6 @@ static long	nr_entries;			/* Number of range entries */
 static int parse_inode_arg(void)
 {
 	int p;
-	u32 imax;
 	u32 range_begin;
 	u32 range_end;
 	u32 range_temp;
@@ -132,7 +297,6 @@ static int parse_inode_arg(void)
 
 	/* init variables */
 	p = strlen(optarg);
-	imax = p;
 	opt_arg_ptr = optarg;
 	opt_arg_end1 = optarg;
 	opt_arg_end2 = &(optarg[p]);
@@ -227,7 +391,8 @@ static void version(void)
 	ntfs_log_info("Copyright (c) 2002-2005 Richard Russon\n"
 			"Copyright (c) 2004-2005 Holger Ohmacht\n"
 			"Copyright (c) 2005      Anton Altaparmakov\n"
-			"Copyright (c) 2007      Yura Pakhuchiy\n");
+			"Copyright (c) 2007      Yura Pakhuchiy\n"
+			"Copyright (c) 2013      Jean-Pierre Andre\n");
 	ntfs_log_info("\n%s\n%s%s\n", ntfs_gpl, ntfs_bugs, ntfs_home);
 }
 
@@ -293,7 +458,10 @@ static void usage(void)
 static int transform(const char *pattern, char **regex)
 {
 	char *result;
-	int length, i, j;
+	int length, i;
+#ifdef HAVE_REGEX_H
+	int j;
+#endif
 
 	if (!pattern || !regex)
 		return 0;
@@ -315,6 +483,7 @@ static int transform(const char *pattern, char **regex)
 		return 0;
 	}
 
+#ifdef HAVE_REGEX_H
 	result[0] = '^';
 
 	for (i = 0, j = 1; pattern[i]; i++, j++) {
@@ -337,6 +506,9 @@ static int transform(const char *pattern, char **regex)
 	result[j+1] = 0;
 	ntfs_log_debug("Pattern '%s' replaced with regex '%s'.\n", pattern,
 			result);
+#else
+	strcpy(result, pattern);
+#endif
 
 	*regex = result;
 	return 1;
@@ -720,13 +892,14 @@ static int parse_options(int argc, char *argv[])
  */
 static void free_file(struct ufile *file)
 {
-	struct list_head *item, *tmp;
+	struct ntfs_list_head *item, *tmp;
 
 	if (!file)
 		return;
 
-	list_for_each_safe(item, tmp, &file->name) { /* List of filenames */
-		struct filename *f = list_entry(item, struct filename, list);
+	ntfs_list_for_each_safe(item, tmp, &file->name) {
+		/* List of filenames */
+		struct filename *f = ntfs_list_entry(item, struct filename, list);
 		ntfs_log_debug("freeing filename '%s'", f->name ? f->name :
 				NONE);
 		if (f->name)
@@ -740,8 +913,9 @@ static void free_file(struct ufile *file)
 		free(f);
 	}
 
-	list_for_each_safe(item, tmp, &file->data) { /* List of data streams */
-		struct data *d = list_entry(item, struct data, list);
+	ntfs_list_for_each_safe(item, tmp, &file->data) {
+		/* List of data streams */
+		struct data *d = ntfs_list_entry(item, struct data, list);
 		ntfs_log_debug("Freeing data stream '%s'.\n", d->name ?
 				d->name : UNNAMED);
 		if (d->name)
@@ -874,6 +1048,74 @@ static void get_parent_name(struct filename* name, ntfs_volume* vol)
 	return;
 }
 
+/*
+ *		Rescue the last deleted name of a file
+ *
+ *	Under some conditions, when a name is deleted and the MFT
+ *	record is shifted to reclaim the space, the name is still
+ *	present beyond the end of record.
+ *
+ *	For this to be possible, the data record has to be small (less
+ *	than 80 bytes), and there must be no other attributes.
+ *	So only the names of plain unfragmented files can be rescued.
+ *
+ *	Returns NULL when the name cannot be recovered.
+ */
+
+static struct filename *rescue_name(MFT_RECORD *mft, ntfs_attr_search_ctx *ctx)
+{
+	ATTR_RECORD *rec;
+	struct filename *name;
+	int off_name;
+	int length;
+	int type;
+
+	name = (struct filename*)NULL;
+	ntfs_attr_reinit_search_ctx(ctx);
+	rec = find_attribute(AT_DATA, ctx);
+	if (rec) {
+		/*
+		 * If the data attribute replaced the name attribute,
+		 * the name itself is at offset 0x58 from the data attr.
+		 * First be sure this location is within the unused part
+		 * of the MFT record, then make extra checks.
+		 */
+		off_name = (long)rec - (long)mft + 0x58;
+		if ((off_name >= (int)le32_to_cpu(mft->bytes_in_use))
+		    && ((off_name + 4)
+				 <= (int)le32_to_cpu(mft->bytes_allocated))) {
+			length = *((char*)mft + off_name);
+			type = *((char*)mft + off_name + 1);
+			/* check whether the name is fully allocated */
+			if ((type <= 3)
+			   && (length > 0)
+			   && ((off_name + 2*length + 2) 
+				<= (int)le32_to_cpu(mft->bytes_allocated))) {
+				/* create a (partial) name record */
+				name = (struct filename*)
+						ntfs_calloc(sizeof(*name));
+				if (name) {
+					name->uname = (ntfschar*)
+						((char*)mft + off_name + 2);
+					name->uname_len = length;
+					name->name_space = type;
+					if (ntfs_ucstombs(name->uname, length,
+							&name->name, 0) < 0) {
+						free(name);
+						name = (struct filename*)NULL;
+					}
+				}
+			if (name && name->name)
+				ntfs_log_verbose("Recovered file name %s\n",
+						name->name);
+			}
+		}
+	}
+	return (name);
+}
+
+
+
 /**
  * get_filenames - Read an MFT Record's $FILENAME attributes
  * @file:  The file object to work with
@@ -956,10 +1198,38 @@ static int get_filenames(struct ufile *file, ntfs_volume* vol)
 		file->max_size = max(file->max_size, name->size_alloc);
 		file->max_size = max(file->max_size, name->size_data);
 
-		list_add_tail(&name->list, &file->name);
+		ntfs_list_add_tail(&name->list, &file->name);
 		count++;
 	}
 
+	if (!count) {
+		name = rescue_name(file->mft,ctx);
+		if (name) {
+			/* a name was recovered, get missing attributes */
+			file->pref_name = name->name;
+			ntfs_attr_reinit_search_ctx(ctx);
+			rec = find_attribute(AT_STANDARD_INFORMATION, ctx);
+			if (rec) {
+				attr = (FILE_NAME_ATTR *)((char *)rec +
+						le16_to_cpu(rec->value_offset));
+				name->flags      = attr->file_attributes;
+
+				name->date_c     = ntfs2timespec(attr->creation_time).tv_sec;
+				name->date_a     = ntfs2timespec(attr->last_data_change_time).tv_sec;
+				name->date_m     = ntfs2timespec(attr->last_mft_change_time).tv_sec;
+				name->date_r     = ntfs2timespec(attr->last_access_time).tv_sec;
+			}
+			rec = find_attribute(AT_DATA, ctx);
+			if (rec) {
+				attr = (FILE_NAME_ATTR *)((char *)rec +
+						le16_to_cpu(rec->value_offset));
+				name->size_alloc = sle64_to_cpu(attr->allocated_size);
+				name->size_data  = sle64_to_cpu(attr->data_size);
+			}
+		ntfs_list_add_tail(&name->list, &file->name);
+		count++;
+		}
+	}
 	ntfs_attr_put_search_ctx(ctx);
 	ntfs_log_debug("File has %d names.\n", count);
 	return count;
@@ -1038,7 +1308,7 @@ static int get_data(struct ufile *file, ntfs_volume *vol)
 		file->max_size = max(file->max_size, data->size_data);
 		file->max_size = max(file->max_size, data->size_init);
 
-		list_add_tail(&data->list, &file->data);
+		ntfs_list_add_tail(&data->list, &file->data);
 		count++;
 	}
 
@@ -1063,6 +1333,7 @@ static struct ufile * read_record(ntfs_volume *vol, long long record)
 	ATTR_RECORD *attr10, *attr20, *attr90;
 	struct ufile *file;
 	ntfs_attr *mft;
+	u32 log_levels;
 
 	if (!vol)
 		return NULL;
@@ -1073,8 +1344,8 @@ static struct ufile * read_record(ntfs_volume *vol, long long record)
 		return NULL;
 	}
 
-	INIT_LIST_HEAD(&file->name);
-	INIT_LIST_HEAD(&file->data);
+	NTFS_INIT_LIST_HEAD(&file->name);
+	NTFS_INIT_LIST_HEAD(&file->data);
 	file->inode = record;
 
 	file->mft = malloc(vol->mft_record_size);
@@ -1101,6 +1372,8 @@ static struct ufile * read_record(ntfs_volume *vol, long long record)
 	ntfs_attr_close(mft);
 	mft = NULL;
 
+	/* disable errors logging, while examining suspicious records */
+	log_levels = ntfs_log_clear_levels(NTFS_LOG_LEVEL_PERROR);
 	attr10 = find_first_attribute(AT_STANDARD_INFORMATION,	file->mft);
 	attr20 = find_first_attribute(AT_ATTRIBUTE_LIST,	file->mft);
 	attr90 = find_first_attribute(AT_INDEX_ROOT,		file->mft);
@@ -1125,6 +1398,8 @@ static struct ufile * read_record(ntfs_volume *vol, long long record)
 	if (get_data(file, vol) < 0) {
 		ntfs_log_error("ERROR: Couldn't get data streams.\n");
 	}
+	/* restore errors logging */
+	ntfs_log_set_levels(log_levels);
 
 	return file;
 }
@@ -1154,7 +1429,7 @@ static struct ufile * read_record(ntfs_volume *vol, long long record)
 static int calc_percentage(struct ufile *file, ntfs_volume *vol)
 {
 	runlist_element *rl = NULL;
-	struct list_head *pos;
+	struct ntfs_list_head *pos;
 	struct data *data;
 	long long i, j;
 	long long start, end;
@@ -1169,13 +1444,13 @@ static int calc_percentage(struct ufile *file, ntfs_volume *vol)
 		return 0;
 	}
 
-	if (list_empty(&file->data)) {
+	if (ntfs_list_empty(&file->data)) {
 		ntfs_log_verbose("File has no data streams.\n");
 		return 0;
 	}
 
-	list_for_each(pos, &file->data) {
-		data  = list_entry(pos, struct data, list);
+	ntfs_list_for_each(pos, &file->data) {
+		data  = ntfs_list_entry(pos, struct data, list);
 		clusters_inuse = 0;
 		clusters_free  = 0;
 
@@ -1279,8 +1554,7 @@ static int calc_percentage(struct ufile *file, ntfs_volume *vol)
 static void dump_record(struct ufile *file)
 {
 	char buffer[20];
-	const char *name;
-	struct list_head *item;
+	struct ntfs_list_head *item;
 	int i;
 
 	if (!file)
@@ -1294,13 +1568,9 @@ static void dump_record(struct ufile *file)
 	if (file->attr_list)
 		ntfs_log_quiet("Metadata may span more than one MFT record\n");
 
-	list_for_each(item, &file->name) {
-		struct filename *f = list_entry(item, struct filename, list);
-
-		if (f->name)
-			name = f->name;
-		else
-			name = NONE;
+	ntfs_list_for_each(item, &file->name) {
+		struct filename *f =
+			ntfs_list_entry(item, struct filename, list);
 
 		ntfs_log_quiet("Filename: (%d) %s\n", f->name_space, f->name);
 		ntfs_log_quiet("File Flags: ");
@@ -1347,8 +1617,8 @@ static void dump_record(struct ufile *file)
 	}
 
 	ntfs_log_quiet("Data Streams:\n");
-	list_for_each(item, &file->data) {
-		struct data *d = list_entry(item, struct data, list);
+	ntfs_list_for_each(item, &file->data) {
+		struct data *d = ntfs_list_entry(item, struct data, list);
 		ntfs_log_quiet("Name: %s\n", (d->name) ? d->name : UNNAMED);
 		ntfs_log_quiet("Flags: ");
 		if (d->resident)   ntfs_log_quiet("Resident\n");
@@ -1388,7 +1658,7 @@ static void dump_record(struct ufile *file)
  *
  * Print a one line description of a file.
  *
- *   Inode    Flags  %age  Date            Size  Filename
+ *   Inode    Flags  %age     Date  Time        Size  Filename
  *
  * The output will contain the file's inode number (MFT Record), some flags,
  * the percentage of the file that is recoverable, the last modification date,
@@ -1408,14 +1678,14 @@ static void dump_record(struct ufile *file)
 static void list_record(struct ufile *file)
 {
 	char buffer[20];
-	struct list_head *item;
+	struct ntfs_list_head *item;
 	const char *name = NULL;
 	long long size = 0;
 	int percent = 0;
 
 	char flagd = '.', flagr = '.', flagc = '.', flagx = '.';
 
-	strftime(buffer, sizeof(buffer), "%F", localtime(&file->date));
+	strftime(buffer, sizeof(buffer), "%F %R", localtime(&file->date));
 
 	if (file->attr_list)
 		flagx = '!';
@@ -1425,8 +1695,8 @@ static void list_record(struct ufile *file)
 	else
 		flagd = 'F';
 
-	list_for_each(item, &file->data) {
-		struct data *d = list_entry(item, struct data, list);
+	ntfs_list_for_each(item, &file->data) {
+		struct data *d = ntfs_list_entry(item, struct data, list);
 
 		if (!d->name) {
 			if (d->resident)
@@ -1469,18 +1739,23 @@ static void list_record(struct ufile *file)
  */
 static int name_match(regex_t *re, struct ufile *file)
 {
-	struct list_head *item;
+	struct ntfs_list_head *item;
 	int result;
 
 	if (!re || !file)
 		return 0;
 
-	list_for_each(item, &file->name) {
-		struct filename *f = list_entry(item, struct filename, list);
+	ntfs_list_for_each(item, &file->name) {
+		struct filename *f =
+			ntfs_list_entry(item, struct filename, list);
 
 		if (!f->name)
 			continue;
+#ifdef HAVE_REGEX_H
 		result = regexec(re, f->name, 0, NULL, 0);
+#else
+		result = regexec(re, f->uname, f->uname_len, NULL, 0);
+#endif
 		if (result < 0) {
 			ntfs_log_perror("Couldn't compare filename with regex");
 			return 0;
@@ -1594,6 +1869,9 @@ static int open_file(const char *pathname)
 		flags = O_RDWR | O_CREAT | O_TRUNC;
 	else
 		flags = O_RDWR | O_CREAT | O_EXCL;
+#ifdef HAVE_WINDOWS_H
+	flags ^= O_BINARY | O_RDWR | O_WRONLY;
+#endif
 
 	return open(pathname, flags, S_IRUSR | S_IWUSR);
 }
@@ -1655,7 +1933,7 @@ static int undelete_file(ntfs_volume *vol, long long inode)
 	int i, j;
 	long long start, end;
 	runlist_element *rl;
-	struct list_head *item;
+	struct ntfs_list_head *item;
 	int fd = -1;
 	long long k;
 	int result = 0;
@@ -1706,18 +1984,25 @@ static int undelete_file(ntfs_volume *vol, long long inode)
 		goto free;
 	}
 
-	if (list_empty(&file->data)) {
+	if (ntfs_list_empty(&file->data)) {
 		ntfs_log_quiet("File has no data.  There is nothing to recover.\n");
 		goto free;
 	}
 
-	list_for_each(item, &file->data) {
-		struct data *d = list_entry(item, struct data, list);
+	ntfs_list_for_each(item, &file->data) {
+		struct data *d = ntfs_list_entry(item, struct data, list);
+		char defname[sizeof(UNKNOWN) + 25];
 
 		if (opts.output)
-				name = opts.output;
+			name = opts.output;
 		else
+			if (file->pref_name)
 				name = file->pref_name;
+			else {
+				sprintf(defname,"%s%lld",UNKNOWN,
+						(long long)file->inode);
+				name = defname;
+			}
 
 		create_pathname(opts.dest, name, d->name, pathname, sizeof(pathname));
 		if (d->resident) {
@@ -1792,7 +2077,7 @@ static int undelete_file(ntfs_volume *vol, long long inode)
 				if (rl[i].lcn == LCN_HOLE) {
 					ntfs_log_verbose("File has a sparse section.\n");
 					memset(buffer, 0, bufsize);
-					for (k = 0; k < rl[k].length * vol->cluster_size; k += bufsize) {
+					for (k = 0; k < rl[i].length * vol->cluster_size; k += bufsize) {
 						if (write_data(fd, buffer, bufsize) < bufsize) {
 							ntfs_log_perror("Write failed");
 							close(fd);
@@ -1895,7 +2180,8 @@ static int scan_disk(ntfs_volume *vol)
 	ntfs_attr *attr;
 	long long size;
 	long long bmpsize;
-	int i, j, k, b;
+	long long i;
+	int j, k, b;
 	int percent;
 	struct ufile *file;
 	regex_t re;
@@ -1908,6 +2194,7 @@ static int scan_disk(ntfs_volume *vol)
 		ntfs_log_perror("ERROR: Couldn't open $MFT/$BITMAP");
 		return -1;
 	}
+	NVolSetNoFixupWarn(vol);
 	bmpsize = attr->initialized_size;
 
 	buffer = malloc(BUFSIZE);
@@ -1926,13 +2213,17 @@ static int scan_disk(ntfs_volume *vol)
 			ntfs_log_error("ERROR: Couldn't create a regex.\n");
 			goto out;
 		}
+#ifndef HAVE_REGEX_H
+		re->upcase = vol->upcase;
+		re->upcase_len = vol->upcase_len;
+#endif
 	}
 
 	nr_mft_records = vol->mft_na->initialized_size >>
 			vol->mft_record_size_bits;
 
-	ntfs_log_quiet("Inode    Flags  %%age  Date           Size  Filename\n");
-	ntfs_log_quiet("---------------------------------------------------------------\n");
+	ntfs_log_quiet("Inode    Flags  %%age     Date    Time       Size  Filename\n");
+	ntfs_log_quiet("-----------------------------------------------------------------------\n");
 	for (i = 0; i < bmpsize; i += BUFSIZE) {
 		long long read_count = min((bmpsize - i), BUFSIZE);
 		size = ntfs_attr_pread(attr, i, read_count, buffer);
@@ -1948,7 +2239,8 @@ static int scan_disk(ntfs_volume *vol)
 					continue;
 				file = read_record(vol, (i+j)*8+k);
 				if (!file) {
-					ntfs_log_error("Couldn't read MFT Record %d.\n", (i+j)*8+k);
+					ntfs_log_error("Couldn't read MFT Record %lld.\n",
+							(long long)(i+j)*8+k);
 					continue;
 				}
 
@@ -1994,6 +2286,7 @@ out:
 	if (opts.match)
 		regfree(&re);
 	free(buffer);
+	NVolClearNoFixupWarn(vol);
 	if (attr)
 		ntfs_attr_close(attr);
 	return results;
@@ -2154,8 +2447,8 @@ int main(int argc, char *argv[])
 
 	utils_set_locale();
 
-	vol = utils_mount_volume(opts.device, MS_RDONLY |
-			(opts.force ? MS_RECOVER : 0));
+	vol = utils_mount_volume(opts.device, NTFS_MNT_RDONLY |
+			(opts.force ? NTFS_MNT_RECOVER : 0));
 	if (!vol)
 		return 1;
 
